from dbmodel import Doc,Cit,Keyword,db
from peewee import *
from collections import Counter

from pathlib import Path
from hashlib import md5
import re

debug=False







def fixcitedauth(a):
    a = a.strip()
    if not len(a):
        return None

    sp = a.lower().split()
    if len(sp) < 2:
        return None
    if len(sp) >= 5:
        return None
    
    l, f = a.lower().split()[:2] # take first two words
    
    if len(l) == 1: # sometimes last and first name are switched for some reason
        l, f = f, l
        
    f = f[0] + "." # first initial
    
    a = ", ".join([l, f]) # first initial
    a = a.title() # title format, so I don't have to worry later
    
    if debug:
        print('cited author:', a)
        
    return a


def generate_references(refs,mydoc, **kwargs):
    cits = []
    for r in refs:
        yspl = re.split("((?:18|19|20)[0-9]{2})", r)

        if len(yspl) < 2:
            continue

        auth, year = yspl[:2]
        auth = auth.strip()
        
        if len(auth.split(" ")) > 5:
            continue
        
        year = int(year)

        if auth == "":
            continue

        auth = fixcitedauth( auth )
        if auth is None: # catching non-people, and not counting the citations
            continue
        
        full_ref = r
        
        if 'DOI' not in full_ref and not ( # it's a book!
            len(re.findall(r', [Vv][0-9]+', full_ref)) or
            len(re.findall(r'[0-9]+, [Pp]?[0-9]+', full_ref))
        ):
            #full_ref = re.sub(r', [Pp][0-9]+', '', full_ref) # get rid of page numbers!
            
            full_ref = "|".join( # splits off the author and year, and takes until the next comma
                [auth]+
                [x.strip().lower() for x in full_ref.split(",")[2:3]]
            )
        else: # it's an article!
            # just adds a fixed name and date to the front
            full_ref = "|".join(
                [auth, str(year)] +
                [",".join( x.strip() for x in full_ref.lower().split(",")[2:] ).split(",doi")[0]]
                )
                
        if debug:
            print('fix_refs_worked',auth,year,full_ref)
            
        c = Cit(
                doc = mydoc,
                author = auth,
                year = year,
                full_ref = full_ref
            )
        cits.append(c)
            
    Cit.bulk_create(cits, batch_size=100)
    for c in cits:
        yield c



def fixcitingauth(au):
    authors = au.split(";")
    for x in authors:
        x = x.strip().lower()
        x = re.sub("[^a-zA-Z\s,]+", "", x) # only letters and spaces allowed

        xsp = x.split(", ")
        if len(xsp) < 2:
            return xsp[0]
        elif len(xsp) > 2:
            print("Warning:", 'author with too many commas', x)
            continue
            raise Exception("author with too many commas", x)

        f, l = xsp[1], xsp[0]
        f = f[0] # take only first initial of first name

        yield "%s, %s" % (l,f)





def generate_documents(basedir):
    import sys
    import csv
    maxInt = sys.maxsize

    basedir = Path(basedir)

    while True:
        # decrease the maxInt value by factor 10
        # as long as the OverflowError occurs.

        try:
            csv.field_size_limit(maxInt)
            break
        except OverflowError:
            maxInt = int(maxInt/10)

    dcount=0
    for i, path_to_data in enumerate( list(basedir.glob("**/*.txt")) ):
        if not( i < 50 ):
            continue
        print('source_file_%s: %s'%(i,path_to_data))

        from csv import DictReader


        with path_to_data.open('r', encoding='utf8', newline='') as in_f:
            docs = []
            kws = []
            for r in DictReader(in_f, delimiter='\t', quoting=3):
                auths = list(fixcitingauth(r['AU']))
                if not len(auths):
                    continue

                try:
                    year = int(r['PY'])
                except ValueError:
                    if r['EA'].strip() == '':
                        print("\n\n\n")
                        print(path_to_data)
                        print('ERR','year',r['PY'],'cnt',r['NR'],r['TI'])
                        continue
                        
                    year = int( r['EA'].split("/")[-1].split(" ")[-1] )

                doc = Doc.create(
                    title = r['TI'],
                    first_author = auths[0],
                    journal = r['SO'].lower().strip(),
                    year = year,
                    citcnt_external = int(r['NR'])
                )
                docs.append(doc)


                for kw in r['ID'].split(";"):
                    kw = kw.strip().lower()
                    if kw == '':
                        continue

                    kws.append(Keyword(
                        doc=doc,
                        text=kw
                    ))

                for ref in generate_references(r['CR'].split(';'), doc):
                    pass
                
                yield doc

            with db.atomic():
                Keyword.bulk_create(kws, batch_size=100)
           

def txt_to_rows( folder ):
    from pathlib import Path

    #wos_base = "path/to/wos/data"
    basedir = Path(folder)
    # ensures there are not overflow errors while importing large CSVs


    counter = 0
    for doc in generate_documents(folder):
        counter += 1

        if counter % 500 == 0:
            print('on doc %s' % counter)


from .. import re
from ..exceptions import invalid_entry

class wos_doc:

    def __init__(self, raw):
        self.raw = raw

        self.journal = raw['SO'].lower()
        try:
            self.publish = int(raw['PY'])
        except ValueError:
            raise invalid_entry("doc year not valid")

        self.citing_authors = list(self.get_citing_authors())
        if not len(self.citing_authors):
            raise invalid_entry("No authors", self.raw['AU'].split(";"))

        self.uid = self.raw['UT']


    def generate_references(self):
        refs = self.raw['CR'].split(';')
        for r in refs:


            try:
                ref = wos_ref(r)

                if "*" in ref.author:
                    continue

                yield ref

            except invalid_entry:
                pass

    def get_citing_authors(self):
        authors = self.raw['AU'].split(";")
        for x in authors:
            x = x.strip().lower()
            x = re.sub("[^a-zA-Z\s,]+", "", x)  # only letters and spaces allowed

            xsp = x.split(", ")
            if len(xsp) != 2:
                yield x
                continue

            f, l = xsp[1], xsp[0]
            f = f[0]  # take only first initial of first name

            yield "%s, %s" % (l, f)


class wos_ref:

    def fixcitedauth(self):
        a = self.author.strip()
        if not len(a):
            raise invalid_entry("author too short")

        sp = a.lower().split()
        if len(sp) < 2:
            raise invalid_entry("author too short")
        if len(sp) >= 5:
            raise invalid_entry("author too long")

        l, f = a.lower().split()[:2]  # take first two words

        if len(l) == 1:  # sometimes last and first name are switched for some reason
            l, f = f, l

        f = f[0] + "."  # first initial

        a = ", ".join([l, f])  # first initial
        a = a.title()  # title format, so I don't have to worry later

        self.author = a

    def __init__(self, raw):




        yspl = re.split("((?:18|19|20)[0-9]{2})", raw)

        if len(yspl) < 2:
            raise invalid_entry("ref too short")

        self.author, self.publish = yspl[:2]
        self.fixcitedauth()

        self.publish = int(self.publish)

        self.full_ref = raw

        if 'DOI' not in self.full_ref and not (  # it's a book!
                len(re.findall(r', [Vv][0-9]+', self.full_ref)) or
                len(re.findall(r'[0-9]+, [Pp]?[0-9]+', self.full_ref))
        ):
            # full_ref = re.sub(r', [Pp][0-9]+', '', full_ref) # get rid of page numbers!

            self.full_ref = "|".join(  # splits off the author and year, and takes until the next comma
                [self.author] +
                [x.strip().lower() for x in self.full_ref.split(",")[2:3]]
            )
            self.type = 'book'

        else:  # it's an article!
            # just adds a fixed name and date to the front
            self.full_ref = "|".join(
                [self.author, str(self.publish)] +
                [",".join(x.strip() for x in self.full_ref.lower().split(",")[2:]).split(",doi")[0]]
            )
            self.type = 'article'


from .. import env, defaultdict, pd

import pickle
from .. import Path, Counter, defaultdict, download_file, VariableNotFound

__all__ = [
    'Dataset','CitedRef'
]

"""
import sys
#sys.path.append('C:/Users/amcga/GitHub/knowknow/archive')
sys.path.append('C:/Users/amcga/GitHub/knowknow')
sys.path = sys.path[1:]
print(sys.path)
from knowknow.datastore_cnts.count_cache import *
wos = Dataset('sociology-wos')
wos(fy=1993)
"""

class CitedRef:
    def __init__(self, identifier):
        self.identifier = identifier

        isplit = self.identifier.split("|")

        self.full_ref = self.identifier
        self.author = isplit[0]
        self.pyear = None
        if len(isplit) == 2:
            self.type = 'book'
            self.title = isplit[1]
        else:
            self.type = 'article'
            self.pyear = int(isplit[1])
            self.title = isplit[2]

    def __repr__(self):
        return "< %s >" % "Cited %s: %s" % (self.type, self.full_ref)

class ent:
    citing_author = 'fa'
    cited_author = 'ta'
    publication_year = 'fy'
    cited_year = 'ty'
    cited_pub = 'c'
    citing_journal = 'fj'
    first_citing_author = 'ffa'

    #@classmethod
    #def c_fn(cls, x):
    #    return CitedRef(x)

    ents = 'fa,ta,fy,ty,c,fj,ffa'.split(",")



class Constants:

    DEFAULT_KEYS = ['fj.fy', 'fy']  # , 'c', 'c.fy']
    data_files = {
        'sociology-wos': 'https://files.osf.io/v1/resources/9vx4y/providers/osfstorage/'
                         '5eded795c67d30014e1f3714/?zip=',
        'sociology-jstor': 'https://files.osf.io/v1/resources/9vx4y/providers/osfstorage/'
                           '5eded76fc67d3001491f27d1/?zip=',
        'sociology-jstor-basicall': 'https://files.osf.io/v1/resources/9vx4y/providers/osfstorage/'
                                   '5eded7478b542601748b4bdb/?zip='
    }
    data_files = {}


save_nametuples = {}
def make_cross(*args, **kwargs):
    global save_nametuples
    from collections import namedtuple
    
    if len(args):
        assert(type(args[0]) == dict)
        return make_cross(**args[0])
    
    keys = tuple(sorted(kwargs))
    
    if keys in save_nametuples:
        my_named_tuple = save_nametuples[keys]
    else:
        my_named_tuple = namedtuple("_".join(keys), keys)
        save_nametuples[keys] = my_named_tuple

    return my_named_tuple(**kwargs)


def named_tupelize(d,ctype):
    keys = sorted(ctype.split("."))
    
    def doit(k):
        if type(k) in [tuple, list]:
            return make_cross(dict(zip(keys, k)))
        elif len(keys) == 1:
            return make_cross({keys[0]:k})
        else: 
            raise Exception("strange case...")
    
    return {
        doit(k):v
        for k,v in d.items()
    }


class Determ:
    def __init__(self, dataset, **kwargs):
        self.dataset = dataset
        self.define = kwargs
        self.keys = sorted(self.define.keys())
        self.cnt = Cnt.by(self.dataset, *self.keys)

        self._noneKeys = set(k for k,v in self.define.items() if v is None)
        self._notNoneKeys = set(k for k,v in self.define.items() if v is not None)

    def refine(self, **kwargs):
        new_args = dict(self.define, **kwargs)
        return Determ(self.dataset, **new_args)


    def _gen_count(self, typ):
        """
        `typ` is either 'doc' or 'cit'
        Internal function to abstract some code.
        """

        if typ == 'doc':
            c = self.cnt.docs
        elif typ == 'cit':
            c = self.cnt.cits
        else:
            raise Exception('Internal error; see someone who knows something...')

        if len(self._noneKeys):
            def gen():
                for item, value in c.items():

                    # this block just skips the non-representative version of C
                    if False:
                        if self.dataset.groups is not None and 'c' in self.define: #DELETED
                            if item.c in self.dataset.groups:
                                myg = self.dataset.groups[ item.c ]
                                #print('grouping!', myg, item.c)
                                if item.c != self.dataset.group_reps[ myg ]:
                                    continue

                    bugger = False
                    for k in self._notNoneKeys:
                        #if type(self.define[k]) in {list,set,range}:
                        #    if getattr(item, k) not in self.define[k]:
                        #        bugger=True
                        #        break
                        #else:
                        if getattr(item, k) != self.define[k]:
                            bugger=True
                            break
                    if bugger:
                        continue

                    ret = {
                        k: getattr(item, k)
                        for k in self._noneKeys
                    }

                    ret['count'] = value

                    if False:
                        # this part actually sums over the values for this item.
                        if self.dataset.groups is not None and 'c' in self.define and item.c in self.dataset.groups: #DELETED
                            myg = self.dataset.groups[item.c]

                            to_set = [k for k,g in self.dataset.groups.items() if g == myg]
                            #print(to_set)
                                
                            count = 0
                            for ts in to_set:
                                #print(item)
                                item_search = item.__class__( *item )
                                item_search = item_search._replace(c=ts)
                                #print(item_search)
                                count += c[ item_search ]
                                #print(item_search, count)

                            ret['count'] = count

                    else:
                        ret['count'] = value

                    yield ret

            dd = pd.DataFrame(list(gen()))
            dd = dd.sort_values(sorted(self._noneKeys))
            return dd

        else:

            return c[ make_cross(**self.define) ]

    @property
    def docs(self):
        return self._gen_count('doc')
    @property
    def cits(self):
        return self._gen_count('cit')







def ktrans(kwargs):
    # restructure kwargs if need be
    newkwargs = {}
    for k,v in kwargs.items():
        if type(v) == CitedRef:
            v = v.identifier

        if k not in ent.ents:
            try:
                k = getattr(ent, k)
            except AttributeError:
                raise Exception("%s was not a recognized count entity. Use ent.*" % k)

        newkwargs[k] = v
    return newkwargs






# this might just be putting an unnecessary # of steps between the code and its result!
class Dataset:

    attr_cache = {}

    def __init__(self, name):
        self.name = name

        #try:
        #    self.groups = self.load_variable('groups')
        #except VariableNotFound:
        #    self.groups = None
        #try:
        #    self.group_reps = self.load_variable('group_reps')
        #except VariableNotFound:
        #    self.group_reps = None



        self.groupings = []
        

        try:
            groups = self.load_variable('groups')
            group_reps = self.load_variable('group_reps')
            self.groupings.append(
                ('c', {k: group_reps[v] for k,v in groups.items()})
            )
        except VariableNotFound:
            pass
        

        self.load_attributes()





    def __call__(self, typ=None, kwarg_dict=None, **kwargs):

        if kwarg_dict is not None:
            kwargs = kwarg_dict
            
        det = Determ(
            dataset = self,
            **kwargs
        )
        if typ is None:
            return det
        else:
            return getattr(det,typ)

        return
        assert(typ in {'cit','doc'})

        keys = list(kwargs.keys())
        count = self.by_raw(keys)

        if typ == 'cits':
            count = count.cits
        if typ == 'docs':
            count = count.docs

        none_keys = [k for k,v in kwargs.items() if v is None]
        kwargs = ktrans(kwargs)

        return Determ(
            dataset = self,
            **kwargs
        )




    def search(self, dtype, name):
        if dtype not in {'fa','ta','c'}:
            raise Exception('datatype error')

        assert(type(name) == str)

        return [x for x in self.items(dtype) if name.lower() in x.lower()]






    def save_figure(self, name):
        from .. import plt
        outdir = Path(env.figure_dir)
        if not outdir.exists():
            outdir.mkdir()
        print("Saving to '%s'"%outdir)
        plt.savefig(str(outdir.joinpath("%s.png" % name)), bbox_inches="tight")



    def display_figure(self, x, titles=True):
        from .. import plt, display, HTML, Image
        displayed_normally = set()

        import os
        import time
        from datetime import datetime
            
        search = list(Path(env.figure_dir).glob(x))
        if not len(search):
            raise Exception("What figure!? '%s'" % x)
        
        for f in sorted(search):
            if titles:
                display(HTML("<h1>%s</h1>" % (f.name)))
            display(Image(filename=f))
            displayed_normally.add(f.name)
                
        return displayed_normally







    def exists(self):
        return Path(env.variable_dir).joinpath(self.name).exists()

    def items(self, what):
        kwargs = ktrans({what: None})

        process = lambda x:x
        fn_name = "%s_fn" % what
        if hasattr(ent, fn_name):
            process = getattr(ent, fn_name)

        myl = sorted( Determ(
            dataset = self,
            **kwargs
        ).cits[ what ] )
        myl = list(map(process, myl))
        return myl

    @property
    def start(self):
        return self['RELIABLE_DATA_STARTS_HERE']
        
    @property
    def end(self):
        return self['RELIABLE_DATA_ENDS_HERE']

    def cnt_keys(self):
        avail = Path(env.variable_dir).joinpath(self.name).glob("doc ___ *")
        avail = [x.name for x in avail]
        avail = [x.split("___")[1].strip() for x in avail]
        return sorted(avail)

    def trend(self, dtype, name):
        from .time_trend import TimeTrend
        return TimeTrend(
            dtype = dtype,
            name = name,
            dataset=self
        )
        
    def variables(self):
        avail = Path(env.variable_dir).joinpath(self.name).glob("*")
        avail_cnts = Path(env.variable_dir).joinpath(self.name).glob("* ___ *")

        avail = set(avail).difference(avail_cnts)
        avail = [x.name for x in avail]
        return sorted(avail)


    def load_variable(self, name):

        # go get this collection, if it isn't already downloaded...
        if not Path(env.variable_dir).joinpath(self.name).exists():
            print("collection", self.name, "does not exist...")
            print("attempting to load from OSF")

            if self.name not in Constants.data_files:
                raise VariableNotFound(("no data file", self.name))

            zip_dest = Path(env.variable_dir, "%s.zip" % self.name)
            if not zip_dest.exists():
                download_file(Constants.data_files[self.name], zip_dest)

            print("Extracting...", str(zip_dest))
            import zipfile
            with zipfile.ZipFile(str(zip_dest), 'r') as zip_ref:
                zip_ref.extractall(str(zip_dest.parent.joinpath(self.name)))

        # otherwise load the pickled file and do it!
        try:
            print('loading variable %s/%s from disk' % (self.name,name))
            return pickle.load( Path(env.variable_dir).joinpath(self.name, name).open('rb') )
        except FileNotFoundError:
            raise VariableNotFound((self.name,name))

    def clear_all(self):
        # remove the files
        for f in Path(env.variable_dir).joinpath(self.name).glob("*"):
            f.unlink()

    def save_variable(self, name, val):
        import pickle
        outdir = Path(env.variable_dir).joinpath(self.name)
        outdir.mkdir(exist_ok=True)
        pickle.dump( val, outdir.joinpath(name).open('wb') )

    def load_attributes(self):
        if self.name in self.attr_cache:
            self.docs = self.attr_cache[self.name]

        else:
            try:
                self.docs = self.load_variable('_attributes')
            except VariableNotFound:
                self.docs = {}

            self.attr_cache[self.name] = self.docs

    def __getitem__(self, what):
        return self.docs[what]
        
    def set_attribute(self, what, value):
        self.docs[what] = value
        self.save_variable('_attributes', self.docs)

    def add_grouping(self, typ, mp):
        self.groupings.append( (typ,mp) )

    def by_raw(self, *args):
        return Cnt.by(self, *args) # passes this dataset as the first argument
        
    def by(self, *args):
        return Cnt.by(self, *args) # passes this dataset as the first argument






class Cnt:
    cache = {}

    def __init__(self, dataset, keys=[]):

        if type(keys) != list:
            keys = [keys]
        self.keys = keys
        
        self.dataset = dataset
        
        self._docs = None
        self._cits = None


    @classmethod
    def by(cls, dataset, *keys):
        if not len(keys):
            raise Exception('by what?')
        if type(keys[0]) == list:
            return cls.by(*keys[0])
            
        keys = list(keys)

        myk = (dataset.name, frozenset(keys))

        if myk not in cls.cache:
            ret = Cnt(dataset, keys)
            cls.cache[myk] = ret

        return cls.cache[myk]
        
    def load_counts(self, varname, k):
    
        ret = defaultdict(int)
        # populate _cits, keeping grouping in mind
        basic_counts = named_tupelize( dict(self.dataset.load_variable(varname)), k )
        for k,v in basic_counts.items():
            
            # replacements by grouping
            #salty_element = False

            for typ, mp in self.dataset.groupings:
                #print(k)
                if hasattr(k, typ):
                    #print(typ, len(mp))
                    
                    current_attr = getattr(k, typ)
                    #print(current_attr)
                    if current_attr not in mp:
                        #salty_element = True
                        continue

                    k = k._replace(**{typ: mp[current_attr]})
                    #print(k)
            
            #if salty_element:
            #    continue

            ret[ k ] += v
        
        return ret

    @property
    def cits(self):
        if self._cits is None:
            k = ".".join(sorted(self.keys))
            varname = "ind ___ %s" % k
            self._cits = self.load_counts(varname, k)
        
        return self._cits

    @property
    def docs(self):
        if self._docs is None:
            k = ".".join(sorted(self.keys))
            varname = "doc ___ %s" % k
            self._docs = self.load_counts(varname, k)
        
        return self._docs


    def collapse(self, filt=lambda x:True, on=None):
        new_c = defaultdict(int)
        for item, count in self._docs.items():
            if not filt(item):
                continue

            new_c[on(item)] += count

        return Counter( dict(new_c.items()) )


# db.drop_tables([Doc,Cit,Keyword], safe=True)
# db.create_tables([Doc,Cit,Keyword], safe=True)

print(Doc.select().count(), 'docs already...')

txt_to_rows(
    "G:/My Drive/2020 ORGANISATION/1. PROJECTS/qualitative analysis of literature/110 CITATION ANALYSIS/000 data/sociology wos")

query = (Doc
         .select(Doc.id, fn.COUNT().alias("c"))
         .join(Cit)
         .switch(Doc)
         .group_by(Doc.id)
         )

j = 0
for i, x in enumerate(query):
    if j > 5:
        break

    j += 1

    print(x.id, x.c)

query = (Cit
         .select(fn.COUNT().alias("c"))
         .join(Doc, on=(Cit.doc == Doc.id))
         .switch(Cit)
         .group_by(Cit.full_ref, Doc.year)
         )

q2 = (Cit
      .select(Cit.full_ref, Doc.year, fn.COUNT().alias("c"))
      .join(Doc, on=(Cit.doc == Doc.id))
      .switch(Cit)
      .group_by(Cit.full_ref, Doc.year)
      .where(query >= 2)
      )

print(query.sql())
print(q2.sql())

query = (Cit
         .select(Cit, Doc, fn.COUNT().alias("c"))
         .join(Doc, on=(Cit.doc == Doc.id))
         .switch(Cit)
         .where(Cit.full_ref == 'Bourdieu, P.|distinction')
         .group_by(Doc.year)
         )

query = (Cit
         .select(Cit, Doc, fn.COUNT().alias("c"))
         .join(Doc, on=(Cit.doc == Doc.id))
         .switch(Cit)
         .group_by(Cit.full_ref, Doc.year)
         )

for i, x in enumerate(query):
    if i > 10:
        break
    print(x.full_ref, x.doc.year, x.c)

baseq = (Cit
         .select(Cit, Doc, fn.COUNT().alias("c"))
         .join(Doc, on=(Cit.doc == Doc.id))
         .switch(Cit))

journalq = (Doc
            .select(Doc.journal, fn.COUNT().alias("c"))
            .group_by(Doc.journal)
            )
print("\n".join(list("%s: %s" % (x.journal, x.c) for x in journalq)))
journals = [x.journal for x in journalq]


def get_jy(j, start=1950, end=2020):
    from collections import defaultdict
    jy = (Doc
          .select(Doc.year, fn.COUNT().alias("c"))
          .where(Doc.journal == j)
          .group_by(Doc.year)
          .order_by(Doc.year)
          )
    my_years = defaultdict(int, {
        d.year: d.c
        for d in jy
    })
    return [my_years[yy] for yy in range(start, end + 1)]


def docsByYear(filter=[], start=1950, end=2020):
    from collections import defaultdict
    jy = (Doc
          .select(Doc.year, fn.COUNT().alias("c"))
          .where(Doc.journal == j)
          .group_by(Doc.year)
          .order_by(Doc.year)
          )
    my_years = defaultdict(int, {
        d.year: d.c
        for d in jy
    })
    return [my_years[yy] for yy in range(start, end + 1)]


def get_ajy(a, j, start=1950, end=2020):
    from collections import defaultdict
    jy = (Doc
          .select(Doc.year, fn.COUNT().alias("c"))
          .where(Doc.journal == j)
          .group_by(Doc.year)
          .order_by(Doc.year)
          )
    my_years = defaultdict(int, {
        d.year: d.c
        for d in jy
    })
    return [my_years[yy] for yy in range(start, end + 1)]


from time import time

st = time()

for j in journals:
    (j, get_jy(j))

print("finished", time() - st)


def search_citing_authors(name):
    q = (Doc
         .select(Doc.first_author, fn.COUNT().alias("c"))
         .where(Doc.first_author.contains(name))
         .group_by(Doc.first_author)
         )

    return Counter({x.first_author: x.c for x in q})


from .env import GLOBS, cfile

if __name__ == '__main__':
    import sys
    from pathlib import Path

    args = sys.argv[1:]

    me = Path(__file__).parent

    if args[0] == 'init':

        import yaml
        from os.path import expanduser
        home = expanduser("~")
        default = Path(home).joinpath("knowknow")

        while 1:
            chosen_dir = input("Enter the directory where knowknow will keep data and code (will be created if doesn't exist) <default: %s> : " % default) or str(default)
            chosen_dir = Path(chosen_dir)
            if not chosen_dir.parent.exists():
                ans = None
                while ans not in "ynYN":
                    ans = input("The parent of this directory does not exist. Continue (y/n)?")
                if ans in "nN":
                    continue

            GLOBS['kkdir'] = str(chosen_dir)

            with cfile.open('w') as f:
                yaml.dump(GLOBS, f)

            if not chosen_dir.exists():
                print("Creating directory.")
                chosen_dir.mkdir(parents=True)
            else:
                print("This directory exists.")
            break

        print("Dataset directory updated")

    elif args[0] == 'clone':

        if len(args) != 2:
            raise Exception('`clone` command takes exactly 1 argument')


        if 'kkdir' not in GLOBS:
            raise Exception('You need to call `python -m knowknow init` first...')

        _, name = args

        if name[0] == '@':
            name = name[1:]

        if not( 'github.com' in name ):
            name = "https://github.com/%s" % name

        short_name = name.split("/")[-1]

        print("Cloning '%s' into '<kk>/code/%s' ..." % (name, short_name))

        from git.repo.base import Repo
        fld = Path(GLOBS['kkdir'], 'code')
        if not fld.exists():
            fld.mkdir()
        Repo.clone_from(name, fld.joinpath(short_name))

    elif args[0] == 'start':
        if len(args) != 2:
            raise Exception('`start` command takes exactly 1 argument')

        where = args[1]
        where = where.split("/")[-1]

        dr = Path(GLOBS['kkdir'], 'code', where)

        import os
        os.chdir(dr)
        os.system('jupyter-lab')

    elif args[0] == 'help':
        print("Possible commands are clone (with GitHub url), init (no args), and start (no args)")

    else:
        print("Command not recognized.")
        print("Possible commands are `clone` (with GitHub url), `init` (no args), and `start` (code directory)")


import string_grouper
import editdistance
from .dbquery import *


import pandas as pd

#class Groups:

# the final variable we are constructing
groups = {}

# tracks the last group-id assigned
new_gid = 0


# what are all the cited references??
dd = count_docs(
    by={"ref":Cit.full_ref},
    min_count = 2
)

strings = list(dd['ref'])

len(strings)

def isarticle(x):
    sp = x.split("|")
    if len(sp) < 2:
        return False
    
    try:
        int(sp[1])
        return True
    except ValueError:
        return False

strings = [x for x in strings if '[no title captured]' not in x]
articles = [x for x in strings if isarticle(x)]
books = [x for x in strings if not isarticle(x)]

print("%s articles, %s books to group" % (len(articles), len(books)))

# grouping books

# this cell may take quite a while to run.
# on Intel i7-9700F this runs in about a minute on 185k names.

books_grouped = string_grouper.match_strings(
    pd.Series(books), 
    number_of_processes=8, 
    min_similarity=0.7
)

from collections import defaultdict

books_grouped[(books_grouped.similarity<1-1e-8)].sort_values("similarity")

# for books, we require that the authors are no more than 1 edit from each other
# even after limiting the comparisons necessary, this takes about 20s on Intel i7-9700F

ft = defaultdict(set)

for i,r in books_grouped.iterrows():
    ls = r.left_side
    rs = r.right_side
    
    if ls == rs:
        continue
    
    la = ls.split("|")[0]
    ra = rs.split("|")[0]
    
    if editdistance.eval(la,ra) > 1:
        continue
    
    ft[ls].add(rs)
    ft[rs].add(ls)
    
print("%s books have some connection to others in a group" % len(ft))

# assigns group-ids based on the relational structure derived thus far
# the code propagates ids through the network, assuming transitivity of equality

def traverse(x, gid):
    global groups
    groups[x] = gid
    
    neighbors = ft[x]
    for n in neighbors:
        if n not in groups:
            traverse(n, gid)
      
for i,k in enumerate(books):
    if k in groups:
        continue
        
    traverse(k, new_gid)
    new_gid += 1

len(set(groups.values()))

Counter(gid for x,gid in groups.items() if len(x.split("|"))==2).most_common(10)

# grouping articles

# this cell may take quite a while to run.
# on Intel i7-9700F this runs in five minutes on 234k entries.

articles_grouped = string_grouper.match_strings(
    pd.Series(articles), 
    number_of_processes=8, # decrease this number to 1 or 2 for slower computers or laptops (the fan might start screaming)
    min_similarity=0.8 # the similarity cutoff is tighter for articles than for books
)

articles_grouped[(articles_grouped.similarity<1-1e-8)].sort_values("similarity")

# for articles, we require that the entire citations is only 1 edit apart.
# even after limiting the comparisons necessary, this takes about 20s on Intel i7-9700F

# this cell produces the `ft` variable, which maps from each term to the set of terms equivalent. I.e., `ft[A] = {B1,B2,B3}`

ft = defaultdict(set)

for i,r in articles_grouped.iterrows():
    ls = r.left_side
    rs = r.right_side
    
    if ls == rs:
        continue
    
    la = ls.split("|")[0]
    ra = rs.split("|")[0]
        
    if editdistance.eval(ls,rs) > 2:
        continue
    
    ft[ls].add(rs)
    ft[rs].add(ls)
    #print(ls,"|||",rs)

print("%s articles have some connection to others in a group" % len(ft))

# assigns group-ids based on the relational structure derived thus far
# the code propagates ids through the network, assuming transitivity of equality

def traverse(x, gid):
    global groups
    groups[x] = gid
    
    neighbors = ft[x]
    for n in neighbors:
        if n not in groups:
            traverse(n, gid)

for i,k in enumerate(articles):
    if k in groups:
        continue
        
    traverse(k, new_gid)
    new_gid += 1

# this line will break execution if there aren't as many groups assigned as we have articles and books
assert( len(articles) + len(books) == len(groups) )

len(set(groups.values()))

len(set(groups.values())) - len(articles)

len(set(groups.values())) - len(books) - len(articles)

len(books)

len(articles)

from collections import defaultdict

# for quicker access
counts = {
    x.ref: int(x['count'])
    for i,x in dd.iterrows()
}

group_members = defaultdict(set)
for k,v in groups.items():
    group_members[v].add(k)

def get_reps(groups):

    for k,v in group_members.items():
        ret[k] = max(v, key=lambda x: counts[x])

    return ret

group_reps = get_reps(groups)

updated = 0
for i, cit in enumerate(Cit.select()):
    if cit.full_ref not in groups:
        continue

    grp = groups[cit.full_ref]

    if len(group_members[grp]) < 2:
        continue

    new_name = group_reps[grp]

    if new_name == cit.full_ref:
        continue

    #print('updating %s with %s' % (cit.full_ref, new_name))
    #break
    cit.full_ref = new_name
    cit.author = new_name.split("|")[0]
    cit.save()
    updated += 1

    if updated % 1000 == 0:
        print("Updated %s citations..."%updated)

database_name = 'wos'

# saving the variable for later
save_variable("%s.groups" % database_name, groups)
save_variable("%s.group_reps" % database_name, groups)


from IPython.core.display import display, HTML, Markdown, Image
from collections import Counter, defaultdict
from random import sample, shuffle
from itertools import chain

from tabulate import tabulate
import re

from pathlib import Path
from os.path import dirname, join
import yaml

BASEDIR = dirname(__file__)

from csv import reader as creader

import networkx as nx
import pandas as pd
import seaborn as sns

import matplotlib.pyplot as plt

plt.style.use('seaborn-whitegrid')
import numpy as np

from logzero import logger

from .exceptions import *

from . import env
from .utility import *

# this uses some imports from here...
from .datastore_cnts import *
# from .datastore_sql import *
from .code_sharing import *
from . import viz
# from .meta_notebook import *

# from .knowknow_base import  *
# from .archive import *

if __name__ == '__main__':
    print("MAINING!")


from . import env, Path
import papermill as pm

__all__ = [
    'MetaNB'
]

class MetaNB:

    def __init__(self, name, debug=False):
        self.full_name = name
        self.space, self.short_name = self.full_name.split("/")
        if self.space == '.':
            self.spacefn = env.notebook_dir
        else:
            self.spacefn = env.modules[self.space]
        self.fn = Path(self.spacefn, self.short_name + ".ipynb")

        assert(self.fn.exists())
    
    def __call__(self, **params):
        
        pm.execute_notebook(
            str(self.fn),
            str(self.fn),
            parameters = params,
            nest_asyncio=True
        )


from .dbmodel import Doc,Cit,Keyword,db
from peewee import *
from collections import Counter

from pathlib import Path
from hashlib import md5
import re




__all__ = [
    'FILT', 'BY', 'UTIL',
    'count_docs', 'count_cits',

    'biggest_docs', 'biggest_cits',
    'Cit1', 'Cit2',
    'bib_network',

    'Doc','Cit'
]






class FILT:pass
FILT.big3america = [
    (Doc.journal, [
        'american journal of sociology',
        'american review of sociology',
        "social forces"
    ])
]
FILT.none = []



class BY:pass
BY.year_published = {'y':Doc.year}


class UTIL:pass
UTIL.count_docs = fn.COUNT(Doc.id.distinct()).alias("count")
UTIL.count_cits = fn.COUNT().alias("count")












def filt_it( query, filt ):
    
    for thing in filt:
        if type(thing) == tuple:
            f, where_what = thing
            if type(where_what) == list:
                query = query.where(f.in_(where_what))
            elif type(where_what) == str:
                if where_what[-1] == "*":
                    query = query.where(f.contains(where_what[:-1]))
                else:
                    query = query.where(f == where_what)   
            else:
                raise Exception('what to do with this type?', type(where_what))
        elif type(thing) == Expression:
            query = query.where(thing)
        else:
            raise Exception('what to do with this type?', type(thing))

    return query

def getq_dc( filt=[] ):
    from peewee import Expression
    query = (Doc
        .select()
        .join(Cit, on=(Cit.doc==Doc.id))
    )

    query = filt_it( query, filt )
    return query


def count_docs(**kwargs):
    kwargs['count'] = UTIL.count_docs
    return count_table(**kwargs)
def count_cits(**kwargs):
    kwargs['count'] = UTIL.count_cits
    return count_table(**kwargs)

def count_table( 
    filt=[], #+FILT.big3america #+ [(Cit.author,'Bourdieu, P.')]
    by=BY.year_published,
    start=1950,
    end=2020,
    count=UTIL.count_docs,
    min_count = -1
):


    from collections import defaultdict
    import pandas as pd

    query = getq_dc(filt)

    query = query.select(
        count, 
        *[ v.alias(k) for k,v in by.items()]
    )

    query = (query
        .group_by(*list(by.values()))
        .order_by(*list(by.values()))
    )

    if min_count != -1:
        query = query.having(SQL('count') >= min_count)

    #print(query.sql())
    #print(query.count())

    # do the query
    query = list(query)

    def produce_rows(query):
        for r in query:
            # retrieve the entries from the SQL query
            outr = {}
            for k,v in by.items():
                #print(k,v)
                if v.model.__name__ == 'Doc':
                    outr[k] = getattr(r, k)
                elif v.model.__name__ == 'Cit':
                    outr[k] = getattr(r.cit, k)

            outr['count'] = r.count
            #outr['doc_id'] = r.id
                
            yield outr

    # convert query to DF
    dd = pd.DataFrame(
        list(produce_rows(query))
    )

    #dd = dd.set_index(
    #    list(by.keys())
    #)

    return dd


def get_model_from_row(row, what):
    if what.model.__name__ == 'Cit':
        return getattr(row.cit, what)
    if what.model.__name__ == 'Doc':
        return getattr(row, what)

def biggest_docs(filt=[], what=Doc.first_author):
    from collections import defaultdict
    q = (getq_dc(filt)
        .select(what.alias('what'), UTIL.count_docs)
        .group_by(what)
        .order_by(SQL("count").desc())
        .limit(100)
    )
    
    def jj(x):
        if what.model.__name__ == 'Cit':
            return x.cit.what
        if what.model.__name__ == 'Doc':
            return x.what

    return [
        ( jj(d), d.count)
        for d in q
    ]

def biggest_cits(filt=[], what=Cit.author, N=100):
    from collections import defaultdict

    q = (getq_dc(filt)
        .select(what.alias('what'), UTIL.count_cits)
        .group_by(what)
        .order_by(SQL("count").desc())
        .limit(N)
    )

    def jj(x):
        if what.model.__name__ == 'Cit':
            return x.cit.what
        if what.model.__name__ == 'Doc':
            return x.what

    return [
        ( jj(d), d.count)
        for d in q
    ]








Cit1 = Cit.alias()
Cit2 = Cit.alias()

def bib_network(filt, cited_ent='full_ref', min_cocit=5, limit=None):
    import networkx as nx
    from peewee import Expression
    
    # join it all together
    query = (Cit1
        .select(getattr(Cit1, cited_ent), getattr(Cit2, cited_ent), fn.COUNT().alias("count"))
        .join(Cit2, JOIN.INNER, on=(
            (Cit1.doc == Cit2.doc) & 
            (Cit1.full_ref != Cit2.full_ref) # not the same citations
        ))
        .join(Doc, on=(Cit1.doc == Doc.id))
        .group_by(getattr(Cit1, cited_ent), getattr(Cit2, cited_ent))
    )

    query = filt_it( query, filt )

    if min_cocit is not None:
        query = query.having(SQL('count') >= min_cocit)

    if limit is not None:
        query = query.limit(limit)

    g = nx.Graph()
    for r in query:
        g.add_edge(
            getattr(r, cited_ent),
            getattr(r.cit, cited_ent), 
            weight=r.count
        )
    return g


"""
Utilities
=======================================
Functions that don't need state of knowknow engine.
Mostly are meant to generate empty/autofilled data structures or for type transformation.
"""
from collections import namedtuple
from pathlib import Path
import requests
from logzero import logger




#divide by row sums
#e/e.sum(axis=1)[:,None]

def matrix_normalize(mat, axis=0):
    if axis==0: # rows
        return mat/mat.sum(axis=1)[:,None]
    elif axis==1: #columns
         return (mat.T/mat.T.sum(axis=1)[:,None]).T
    else:
        raise Exception('axis != 0/1')






def memoize(func):
    cache = dict()

    def memoized_func(*args):
        if tuple(args) in cache:
            return cache[tuple(args)]
        result = func(*args)
        cache[tuple(args)] = result
        return result

    return memoized_func


def comb(x, y):
    '''
    function to combine and sort . seprated values
    :param x: . separated values
    :param y: . separated values
    :return:  combined sorted list x+y
    '''
    a = set(x.split("."))
    b = set(y.split("."))

    return ".".join(sorted(a.union(b)))


def make_cross_kwargs(*args, **kwargs):
    """
    Looks like it turns a dictionary into a named tuple.
    TODO: get rid of kwargs logic. Why generating a named tuple is named cross?
    :param args:
    :param kwargs:
    :return:
    """
    # removed caching
    if len(args):
        assert (type(args[0]) == dict)
        return make_cross_kwargs(**args[0])

    keys = tuple(sorted(kwargs))

    my_named_tuple = namedtuple("_".join(keys), keys)

    return my_named_tuple(**kwargs)


@memoize
def gen_tuple_template(keys: tuple):
    """
    :param keys: a tuple of strings
    :return: a named tuple template
    ('e', 'q', 'w') -> <class '__main__.e_q_w'>
    """
    # TODO: add caching
    return namedtuple('_'.join(keys), keys)


def make_cross(key___value: dict):
    """
    :param key___value:
    :return:
    {'1': 'q', '2': 'w', '3': 'e'} -> e_q_w(e='3', q='1', w='2')
    """
    return gen_tuple_template(tuple(sorted(key___value.keys())))(**key___value)


@memoize
def doit(k, keys: tuple):
    if type(k) in [tuple, list]:
        return make_cross(dict(zip(keys, k)))
    elif len(keys) == 1:
        # TODO: where this hack is used?
        # logger.warning(f'DoIt function in named_tupelize uses len(1) hack. ctype: {ctype}, '
        #                f'd: {d}, curr_key {k}')
        return make_cross({keys[0]: k})
    else:
        raise Exception(f"Cannot transform {k} into a named tuple with keys {keys}")


def named_tupelize(d: dict, ctype: str):
    '''
    For each key in dictionary d, try making it a named tuple. Return {namedtuple(k, ctypes): v for k, v in dict)
    :param d: input dictionary
    :param ctype: Ctype
    :return: named tuple
    '''
    keyss = tuple(sorted(ctype.split(".")))
    return {doit(k, keyss): v for k, v in d.items()}


class VariableNotFound(Exception):
    pass


def download_file(url, outfn):
    '''
    Downlaod file from given url
    :param url: url of file to be downloaded
    :param outfn: filename of downloaded file
    :return:
    '''
    url = str(url)
    outfn = str(outfn)
    Path(outfn).parent.mkdir(exist_ok=True)
    logger.info(f"Beginning download, {url}")
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(outfn, 'wb') as f:
            for i, chunk in enumerate(r.iter_content(chunk_size=8192)):
                if i % 1000 == 0 and i:
                    logger.info('%0.2f MB downloaded...' % (i * 8192 / 1e6))
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                # if chunk:
                f.write(chunk)
    return outfn



__all__ = [
    'require'
]

from knowknow.env import GLOBS

modules = {
    'citation-deaths': "G:/My Drive/2020 ORGANISATION/1. PROJECTS/qualitative analysis of literature/110 CITATION ANALYSIS/010 analyses/bundle 101 - citation deaths reboot 10-2020",
    'wos-counter': "G:/My Drive/2020 ORGANISATION/1. PROJECTS/qualitative analysis of literature/110 CITATION ANALYSIS/010 analyses/bundle 102 - creating wos database",
    'stats': "G:/My Drive/2020 ORGANISATION/1. PROJECTS/qualitative analysis of literature/110 CITATION ANALYSIS/010 analyses/bundle 103 - various statistics"
    #'summary-stats': ""
}

#require('amcgail/citation-deaths', '.')

mod_cache = {}

def require(where, what):

    from pathlib import Path

    wsp = where.split("/")
    if len(wsp) == 1:
        who, name = None, wsp[0]
    else:
        who, name = wsp

    dest_fn = Path(GLOBS['kkdir'], 'code', name)

    if not dest_fn.exists():
        print("Module doesn't exist. Attempting to load it...")

        if where[0] == '@':
            where = where[1:]

        url = "https://github.com/%s" % where

        short_name = url.split("/")[-1]

        print("Cloning '%s' into '<kk>/code/%s' ..." % (name, short_name))

        from git.repo.base import Repo
        fld = Path(GLOBS['kkdir'], 'code')
        if not fld.exists():
            fld.mkdir()
        Repo.clone_from(url, fld.joinpath(short_name))


    wsp = what.split("/")
    if len(wsp) == 1:
        # get it from __init__
        modfn = dest_fn.joinpath("__init__.py")

    elif len(wsp) == 2:
        # get it from file within
        modfn = dest_fn.joinpath("%s.py" % wsp[0])
    else:
        raise Exception('yolo')


    if modfn in mod_cache:
        mod = mod_cache[modfn]
    else:
        from runpy import run_path
        mod = run_path( modfn )

        if '__all__' in mod:
            mod = {
                k: v for k,v in mod.items() if k in mod['__all__']
            }

        mod_cache[modfn] = mod

    return mod[wsp[-1]]


import sqlalchemy

def fix_refs(refs):
    for r in refs:
        yspl = re.split("((?:18|19|20)[0-9]{2})", r)

        if len(yspl) < 2:
            continue

        auth, year = yspl[:2]
        auth = auth.strip()
        
        if len(auth.split(" ")) > 5:
            continue
        
        year = int(year)

        if auth == "":
            continue

        auth = fixcitedauth( auth )
        if auth is None: # catching non-people, and not counting the citations
            continue
        
        full_ref = r
        
        if 'DOI' not in full_ref and not ( # it's a book!
            len(re.findall(r', [Vv][0-9]+', full_ref)) or
            len(re.findall(r'[0-9]+, [Pp]?[0-9]+', full_ref))
        ):
            #full_ref = re.sub(r', [Pp][0-9]+', '', full_ref) # get rid of page numbers!
            
            full_ref = "|".join( # splits off the author and year, and takes until the next comma
                [auth]+
                [x.strip().lower() for x in full_ref.split(",")[2:3]]
            )
        else: # it's an article!
            # just adds a fixed name and date to the front
            full_ref = "|".join(
                [auth, str(year)] +
                [",".join( x.strip() for x in full_ref.lower().split(",")[2:] ).split(",doi")[0]]
                )

        if use_included_citations_filter:
            if full_ref not in included_citations:
                continue
            
        # implement grouping of references
        if groups is not None:
            if full_ref in groups:
                # retrieves retrospectively-computed groups
                full_ref = group_reps[
                    groups[full_ref]
                ]
            elif full_ref in ysumc and '[no title captured]' not in full_ref:
                # a small minority, the ones which are dropped in this process anyways
                #print('not in grouping')
                raise Exception(full_ref, "not in grouping!")
            else:
                continue
                
        if debug:
            print('fix_refs_worked',auth,year,full_ref)
        yield ( auth, year, full_ref )
        

def _txt_to_rows( folder ):
    from pathlib import Path

    #wos_base = "path/to/wos/data"
    wos_base = "G:/My Drive/2020 ORGANISATION/1. PROJECTS/2. CURRENT/qualitative analysis of literature/pre 5-12-2020/009 get everything from WOS"
    basedir = Path(wos_base)
    # ensures there are not overflow errors while importing large CSVs

    import sys
    import csv
    maxInt = sys.maxsize

    while True:
        # decrease the maxInt value by factor 10
        # as long as the OverflowError occurs.

        try:
            csv.field_size_limit(maxInt)
            break
        except OverflowError:
            maxInt = int(maxInt/10)

    db_path = 'wos.db'
    engine = sqlalchemy.create_engine(f'sqlite:///{db_path}')

    dcount=0
    for i, path_to_data in enumerate( list(basedir.glob("**/*.txt")) ):
        print('source_file_%s: %s'%(i,path_to_data))

        #path_to_data = '1-500.txt'
        if i == 0:
            behaviour = 'replace'  # fail, append
        else:
            behaviour = 'append'

        from csv import DictReader
        with path_to_data.open('r', encoding='utf8', newline='') as in_f:
            df = pandas.DataFrame(list(DictReader(in_f, delimiter='\t')))
        # df.to_sql('wos', engine)

        articles = df
        articles.to_sql('articles', index=False, con=engine, if_exists=behaviour)

        authors = []
        citations = []
        articles = []
        for idx, row in df.iterrows():
            if pandas.isna(row.AU) or pandas.isna(row.CR):
                continue
            uid = articles['myuid'].loc[idx]
            names = row['AU'].split(';')
            fullnames = row['AF'].split(';')

            names = [x.strip() for x in names]
            fullnames = [x.strip() for x in fullnames]

            if len(names) != len(fullnames):
                print("Skipping bc. name discrepancy")
                continue

            def fixcitingauth():
                authors = r['AU'].split(";")
                for x in authors:
                    x = x.strip().lower()
                    x = re.sub("[^a-zA-Z\s,]+", "", x) # only letters and spaces allowed

                    xsp = x.split(", ")
                    if len(xsp) < 2:
                        return xsp[0]
                    elif len(xsp) > 2:
                        raise Exception("author with too many commas", x)

                    f, l = xsp[1], xsp[0]
                    f = f[0] # take only first initial of first name

                    yield "%s, %s" % (l,f)

            citing_authors = list(fixcitingauth())
            
            curr_authors = pandas.DataFrame({'name': names, 'fullname': fullnames, 'idx': range(len(names))})
            curr_authors['name_key'] = list(fixcitingauth())
            
            authors.append(curr_authors)

            curr_cit = fix_refs( row['CR'].split(';') )
            citations.append(pandas.DataFrame({
                'auth':[x[0] for x in curr_cit],
                'year':[x[1] for x in curr_cit],
                'full_ref':[x[2] for x in curr_cit]
            }))
            citations[-1]['citing_work'] = uid

        authors = pandas.concat(authors)
        citations = pandas.concat(citations)

        authors.to_sql('authors', index=False, con=engine, if_exists=behaviour)
        citations.to_sql('citations', index=False, con=engine, if_exists=behaviour)
        